//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
// $Id: card.hh,v 1.2 2005/01/21 07:15:43 loreti Exp $
//
// Definition of the classes and methods used to abstract a
// single card and a pack of cards, and the operations that we
// need to perform on them.
//
// -------------------------------------------------------------

#ifndef CARDS_HH_
#define CARDS_HH_

#include <iostream>
#include <vector>
#include <algorithm>
#include <cassert>

using std::ostream;
using std::vector;
using std::random_shuffle;


// The class use to represent a card.
//
// The abstraction uses a "suit" and a "value" only; the number
// of suits, and the number of values for every suit, are static
// constants of the class; the internal representation uses a
// suit representation from 0 to card::numSuits-1, and a value
// representation from 1 to card::maxValue (a value of zero is
// used as an "impossible" value).
//
// Other supplied static constants include:
// - card::totCards, the total number of cards, that must of
//   course have the value card::numSuits*card::maxValue;
// - card::suitName and card::valueName, a character
//   representation of the different suits and values (the
//   "impossible" value included).
//
// The supplied methods are: a constructor from a given suit and
// value; a constructor from an unsigned integer interpreted as
// card::maxValue*suit+(value-1) , i.e. having the values from 0
// to card::maxValue-1; two inspectors to get suit and value;
// and a method to print a card on an output stream.

class card {
  unsigned int suit_;
  unsigned int value_;

public:
  static const unsigned int maxValue = 10;
  static const unsigned int numSuits = 4;
  static const unsigned int totCards = 40;
  static const char         suitName[4];
  static const char         valueName[11];

  card(unsigned int sv) {
    suit_  = sv / maxValue;
    value_ = sv % maxValue + 1;
    assert(static_cast<unsigned int>(suit_)  <  numSuits);
    assert(static_cast<unsigned int>(value_) <= maxValue);
  }

  card(unsigned int suit, unsigned int value)
    : suit_(suit), value_(value) {
    assert(static_cast<unsigned int>(suit_)  <  numSuits);
    assert(static_cast<unsigned int>(value_) <= maxValue);
  }

  ~card() { }

  unsigned int getSuit()  const { return suit_;  }
  unsigned int getValue() const { return value_; }

  void out(ostream &) const;
  friend ostream & operator << (ostream & o, const card & c)
    { c.out(o); return o; }
};

// The class use to represent a pack of cards.
//
// A pack of cards is abstracted as a vector of cards; the
// default constructor initializes that vector with
// card::totCards different values.  The other methods supplied
// may: shuffle an existing pack; return a const_iterator th the
// first card; and print the pack content on an output stream.
//
// Copy constructor and assignment operator are declared private
// and never defined, in order to exclude their use.

class pack {
  vector<card> cards_;

  pack(const pack &);
  pack & operator = (const pack &);

public:
  pack();
  ~pack() { }

  vector<card>::const_iterator getCI() const
    { return cards_.begin(); }
  void shuffle()
    { random_shuffle(cards_.begin(), cards_.end()); }
  void print (ostream &) const;
  friend ostream & operator << (ostream & o, const pack & p)
    { p.print(o); return o; }
};

#endif // CARDS_HH_
