//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
// $Id: solve.cxx,v 1.5 2005/01/21 07:15:43 loreti Exp $
//
// The procedures that actually solve the patience.  The gaming
// strategy is described later.
//
// -------------------------------------------------------------


#include <iostream>
#include <algorithm>
#include "table.hh"

using namespace std;

#ifdef STLSORT
static inline bool lessStack(cardStack *, cardStack *);
#endif

unsigned int table::solve()
{

// The main solving loop.  The steps are:
// - move all you can to the foundations ==> in moveOut() ;
// - move some cards on the table stacks ==> in moveIn() ;
// - iterate until or: all the stacks are empty (firstStack
//    reaches 16), or no move is possible (moveIn returns
//    false);
// - count how many cards are still on the stacks and return
//   that number to the caller.

  if (moveOut()) {
    while (firstStack < 16   &&   moveIn()) {
      moveOut();
    }
  }

  unsigned int retval = 0;
  for (int i = 0;   i < 16;   i++) {
    retval += pStack[i]->nCards();
  }
  return retval;
}

bool table::moveOut()
{

// moveOut() moves cards from the stacks to the foundations:
//
// - loops over all the non-empty card stacks checking the first
//   card;
// - when a suitable card is found it is moved out of the stack
//   to the foundation, and the following card on that stack is
//   also checked;
// - all is repeated until no more possible moves are found.
//
// When a stack is emptied in that loop, is moved to the
// beginning of the table (i.e. swapped with the first non-empty
// one).
//
// The return value is "true" is at least a card has been moved.

  int moves = 0;
  int partial;

#ifdef DEBUG
  cout << "--- Entered moveOut() ... ";
#endif

  do {
    partial = 0;
    for (unsigned int i = firstStack;   i < 16;   i++) {
      int suit, value, number;

      for (;;) {
        pStack[i]->look(suit, value, number);
        if (number < 1) {
          if (i != firstStack) swap(pStack[i], pStack[firstStack]);
          firstStack++;
          break;
        }
        if (static_cast<int>( lastValue[suit] ) != --value) {
          break;
        }
        (lastValue[suit])++;
        pStack[i]->remove();
        partial++;
      }
    }
    moves += partial;
  } while (partial);

#ifdef DEBUG
  cout << moves << " cards moved out.\n";
  if (moves) {
    print(cout);
    cout << endl;
  }
#endif

  return (moves > 0);
}

bool table::moveIn()
{
  int i, lastS, lastV;
  int moves[16];
  int *lastM = 0;

// As the first step, the stacks are rearranged in _descending_
// order, according to their first card (always visible, and the
// only movable card of the stack itself).

#ifdef STLSORT
  sort(pStack+firstStack, pStack+16, lessStack);
#else
  tableSort(pStack, firstStack, 15);
#endif

#ifdef DEBUG
  cout << "--- Entered moveIn() ... stacks sorted\n";
  print(cout);
  cout << endl;
#endif

// Then, in moves[i], we build the number of (visible) cards
// that we can move on the top of the (visible) card of the i-th
// stack; taking into account that is meaningless a move from a
// stack containing a single card.
//
// In every stack, "level" (initialized from the constructor) is
// set to reflect "how good" is to move cards on it.  If a stack
// contains only one card, a move don't "trap" any other card;
// and "level" is set to 0 (the lowest possible value, i.e. "to
// move here is an ideal move").
//
// As an aside, if the first visible cards is a 10, it cannot be
// moved other than to the foundations; the underlying cards are
// already trapped.  "level" is set to 0 here too, because
// moving to this stack changes nothing at all.
//
// For all the others situations, "level" is set to the number
// of "trapped" cards; note that moving cards to a stack do not
// raise its "level" (all the moved cards can be removed to the
// foundations is a single step).  But when we remove a card (or
// a sequence of cards) from a stack, we have to lower its
// "level" by one (a covered card is uncovered).

  for (i = firstStack,  lastS = -1,  lastV = -1;   i < 16;   i++) {
    int suit, value, number;

    moves[i] = 0;

    pStack[i]->look(suit, value, number);

    if (value == 10) pStack[i]->level = 0;

    if (number > 1   &&   lastS == suit   &&   lastV == value+1) {
      (*lastM)++;
    } else {
      lastM = moves+i;
    }

    lastS = suit;
    lastV = value;
  }

// Now we find the best move: the move with the lower
// stack[i].level value; and if more than a move is possible
// with the same level, the one with the higher moves[i].
//
// The function returns "true" if (at least) a move is
// possible.

  int bestI = -1;
  for (i = firstStack;   i < 16;   i++) {
    if (moves[i] > 0) {
      if (bestI < 0   ||
          pStack[i]->level < pStack[bestI]->level   ||
          (pStack[i]->level == pStack[bestI]->level &&
           moves[i] > moves[bestI])) {
        bestI = i;
      }
    }
  }

  if (bestI < 0) {

#ifdef DEBUG
    cout << "--- No moves\n";
#endif
    return false;

  } else {

#ifdef DEBUG
    cout << "--- Best move: " << moves[bestI] << " card"
         << (moves[bestI] == 1 ? " on " : "s on ")
         << bestI << endl;
#endif

    for (i = 1;   i <= moves[bestI];   i++) {
      pStack[bestI]->add( pStack[bestI+i]->remove() );
    }
#ifdef DEBUG
    print(cout);
    cout << endl;
#endif
    return true;
  }
}

// Here is the function that sorts the array of cardStack
// pointers in *descending* order.  A few notes:
//
// - there is a 'sort' function in the C++ Standard Library; but
//   its performances are worst then the ones got by an home
//   made procedure.  Probably that sort is not well-suited for
//   my array, small (16 elements at the beginning, less later
//   when some stacks become empty) and almost-sorted after the
//   first call.
//
// - I have experimented with two compilers, three libraries and
//   three sorting methods; the results are reported later.
//   This has been checked on a Linux box: the CPU was Pentium
//   III running 450 MHz with 256 MB RAM; the OS was the kernel
//   2.2.12-20, i.e. RedHat 6.1.  In a different environment,
//   your mileage may vary.
//
// - I used to compile and link:
//   1) Kuck and Associates (KAI) KCC version 3.3g, with the
//      make variable CXXFLAGS set to -O2 -DNDEBUG -DS , where
//      "S" is the sort type;
//   2) g++ 2.95.2 with the default library and CXXFLAGS set to
//      -Wall -O2 -DNDEBUG -DS;
//   3) g++ 2.95.2 with the "version 3" standard library, namely
//      libstdc++ 2.90.8, and CXXFLAGS = -Wall -O2 -DNDEBUG
//      -fhonor-std -DS .
//   and ran with ./sol -t1000000 -p0 , obtaining the following
//   results for the seconds of elapsed CPU time:
//
//                   +-----------------------------------------------+
//                   |                 SORTING METHOD                |
//                   +-----------+-----------+-----------+-----------+
//                   | QUICKSORT | INSERTION | SHELLSORT | C++ LIBR. |
//   +---------------+-----------+-----------+-----------+-----------+
//   | KCC           |   147.94  |   135.96  |   131.88  |   137.79  |
//   +---------------+-----------+-----------+-----------+-----------+
//   | g++ & lib. v2 |   120.73  |   109.44  |   110.37  |   119.46  |
//   +---------------+-----------+-----------+-----------+-----------+
//   | g++ & lib. v3 |   125.61  |   114.09  |   115.19  |   124.53  |
//   +---------------+-----------+-----------+-----------+-----------+

#ifdef QUICKSORT
void table::tableSort (
  cardStack *v[],
  int left,
  int right
) {
  if (left >= right) return;

  int i, last;

  swap(v[left], v[(left+right)/2]);
  last = left;
  for (i = left+1;   i <= right;   i++) {
    if (*(v[i]) > *(v[left])) swap(v[++last], v[i]);
  }
  swap(v[left], v[last]);
  tableSort(v, left,   last-1);
  tableSort(v, last+1, right);
}
#endif // QUICKSORT

#ifdef INSERTIONSORT
void table::tableSort (
  cardStack *v[],
  int left,
  int right
) {
  int i;

  for (i = left + 1;   i <= right;   i++) {
    if (*(v[i]) > *(v[left])) {
      swap(v[left], v[i]);
    }
  }
  for (i = left + 2;   i <= right;   i++) {
    int j = i;
    cardStack *pCS = v[i];

    while (*pCS > *(v[j-1])) {
      v[j] = v[j-1];
      j--;
    }
    v[j] = pCS;
  }
}
#endif // INSERTIONSORT

#ifdef SHELLSORT
void table::tableSort (
  cardStack *v[],
  int left,
  int right
) {
  int h;

  for (h = 1;   h <= (right-left)/9;   h = 3*h+1) { }
  while (h > 0) {
    for (int i = left+h;   i <= right;   i++) {
      int j = i;
      cardStack *pCS = v[i];

      while (j >= left+h   &&   *pCS > *(v[j-h])) {
        v[j] = v[j-h];
        j -= h;
      }
      v[j] = pCS;
    }
    h /= 3;
  }
}
#endif // SHELLSORT

#ifdef STLSORT
static inline bool lessStack(
  cardStack *p1,
  cardStack *p2
) {
  return *p1 > *p2;
}
#endif
